package rdt;

/**
 *  An AlternatingBitProtocol implements the RDTProtocol interface
 *  using an alternating bit algorithm.
 */
public class AlternatingBitProtocol
extends RDTProtocol {

    // send state
    int nextSendSequenceNumber;
    Packet packet0;
    Packet packet1;
    RDTTimer timer;

    // receive state
    int expectedReceiveSequenceNumber;
    Packet ack0;
    Packet ack1;

    /**
     *  rdtp.doInit() is invoked to perform protocol-dependent
     *  initialization for this AlternatingBitProtocol.
     */   
    protected void doInit() {
        // initialize send state
        nextSendSequenceNumber = 0;
        packet0 = createPacket();
        packet0.setType(Packet.DATA_TYPE);
        packet0.setSequenceNumber(0);
        packet1 = createPacket();
        packet1.setType(Packet.DATA_TYPE);
        packet1.setSequenceNumber(1);
        timer = createTimer();
        setAcceptingNewSegments(true);
        setPendingAcknowledgements(false);
        // initialize receive state
        expectedReceiveSequenceNumber = 0;
        ack0 = createPacket();
        ack0.setType(Packet.ACK_TYPE);
        ack0.setSequenceNumber(0);
        ack1 = createPacket();
        ack1.setType(Packet.ACK_TYPE);
        ack1.setSequenceNumber(1);
    }  // protected void doInit()

    /**
     *  handleNewSegment(id) is invoked to handle a new
     *  segment with identification number id.
     *  The identification number is just a stand-in for
     *  the real data in a segment.
     *  It is used by the Logger for recording purposes.
     */   
    public void handleNewSegment(int id) {
        // this is always send side
        setAcceptingNewSegments(false);
        setPendingAcknowledgements(true);
        timer.start();
        if (nextSendSequenceNumber == 0) {
            packet0.setIdentificationNumber(id);
            send(packet0);
        } else if (nextSendSequenceNumber == 1) {
            packet1.setIdentificationNumber(id);
            send(packet1);
        }
    }  // public void handleNewSegment(int)

    /**
     *  handleReceivedData(pkt) is invoked to handle a
     *  DATA_TYPE packet received from its TCPEntity's
     *  connection partner.
     */   
    public void handleReceivedData(Packet pkt) {
        // do receive state modifications
        if (expectedReceiveSequenceNumber == 0) {
            if (pkt.getSequenceNumber() == 0) {
                send(ack0);
                reassemble(pkt.getIdentificationNumber());
                expectedReceiveSequenceNumber = 1;
            } else {
                send(ack1);
            }
        } else if (expectedReceiveSequenceNumber == 1) {
            if (pkt.getSequenceNumber() == 1) {
                send(ack1);
                reassemble(pkt.getIdentificationNumber());
                expectedReceiveSequenceNumber = 0;
            } else {
                send(ack0);
            }
        }
    }  // public void handleReceivedData(Packet)

    /**
     *  rdtp.handleReceivedAck(pkt) is invoked to handle a
     *  ACK_TYPE packet received from its TCPEntity's
     *  connection partner.
     */   
    public void handleReceivedAck(Packet pkt) {
        // do send state modifications
        if (nextSendSequenceNumber == 0) {
            if (pkt.getSequenceNumber() == 0) {
                timer.stop();
                nextSendSequenceNumber = 1;
                setAcceptingNewSegments(true);
                setPendingAcknowledgements(false);
            }
        } else if (nextSendSequenceNumber == 1) {
            if (pkt.getSequenceNumber() == 1) {
                timer.stop();
                nextSendSequenceNumber = 0;
                setAcceptingNewSegments(true);
                setPendingAcknowledgements(false);
            }
        }
    }  // public void handleReceivedAck(Packet)

    /**
     *  rdtp.handleTimedOutPacket(pkt) is invoked to handle a
     *  packet acknowledgement timeout for pkt.
     *  pkt may be null for protocols that use a single timer.
     */   
    public void handleTimedOutPacket(Packet pkt) {
        // this is always send side unless you have a receive side timer
        timer.start();
        if (nextSendSequenceNumber == 0) {
            send(packet0);
        } else if (nextSendSequenceNumber == 1) {
            send(packet1);
        }
    }  // public void handleTimedOutPacket(Packet)

}  // public class AlternatingBitProtocol

