package rdt;

/**
 *  The Internet class provides delivery of Packet objects between
 *  TCPEntity objects.
 *  This abstraction represents the view of the internet from just
 *  above the multiplexer/demultipler in the Transport layer.
 *  <p>
 *  All methods in this class are static.
 *  Messages should be sent to the Internet class object as shown in
 *  method comments.
 */
public class Internet {

    /**
     *  min_rtt is the minimum round-trip time for the Internet
     *  in milliseconds.
     */   
    private static double min_rtt = 100.0;

    /**
     *  max_rtt is the maximum round-trip time for the Internet
     *  in milliseconds.
     */   
    private static double max_rtt = 120.0;

    /**
     *  roundTripTime is the average round-trip time for the Internet
     *  in milliseconds.
     */   
    private static int roundTripTime = 150;

    /**
     *  reliability is probability that a Packet gets delivered
     *  in the Internet.
     */   
    private static double reliability = 0.8;

    /**
     *  random is a random number generator for determining delay
     *  times and deciding whether or not a packet gets delivered.
     */   
    private static final java.util.Random random = new java.util.Random();

    /**
     *  timerFactory is the factory for timers that determine when
     *  packets get delivered.
     */   
    private static final TimerFactory timerFactory =
            new TimerFactory("Internet");

    /**
     *  getDelay returns a randomized delay with average roundTripTime/2.
     */ 
    private static int getDelay() {
        double delay = min_rtt + random.nextDouble()*(max_rtt - min_rtt);
        return Math.round((float)(delay/2.0));
    }  // private int getDelay()

    /**
     *  shouldDeliver returns a boolean whose truth probability is
     *  reliability.
     */ 
    private static boolean shouldDeliver() {
        return random.nextDouble() < reliability;
    }  // private boolean shouldDeliver()

    /**
     *  Internet.printConfiguration() prints the configuration parameters
     *  for the Internet.
     */ 
    public static void printConfiguration() {
        System.out.println("Internet configuration parameters:");
        System.out.println("----------------------------------");
        System.out.println("Minimum round-trip time: " + min_rtt);
        System.out.println("Maximum round-trip time: " + max_rtt);
        System.out.println("Reliability: " + reliability);
        System.out.println();
    }  // public static void printConfiguration()

    /**
     *  Internet.setSeed(sd) sets the seed for the random number generator
     *  used in the Internet to sd.
     *  This method should only be invoked before a simulation run is
     *  started.
     *  <p>
     *  For a given seed, the sequence of delivery successes and delivery
     *  times will always be the same.
     *  If this method is not invoked, the current time will be used to set
     *  the seed, resulting in a different sequence of delivery successes
     *  and delivery times from one run to another.
     */ 
    public static void setSeed(long sd) {
        random.setSeed(sd);
    }  // public static void setSeed(long)

    /**
     *  Internet.getRoundTripTime() returns the average round-trip time for
     *  the Internet.
     */ 
    public static int getRoundTripTime() {
        return Math.round((float)((min_rtt + max_rtt)/2.0));
    }  // public static int getRoundTripTime()

    /**
     *  Internet.setRoundTripTimes(min, max) sets the minimum and maximum
     *  round-trip time for the Internet to min and max.
     */ 
    public static void setRoundTripTimes(int min, int max) {
        min_rtt = min;
        max_rtt = max;
    }  // public static void setRoundTripTimes(int, int)

    /**
     *  Internet.getReliability() returns the probability that a Packet
     *  gets delivered in the Internet.
     */ 
    public static double getReliability() {
        return reliability;
    }  // public static double getReliability()

    /**
     *  Internet.setReliability(rel) sets the probability that a Packet
     *  gets delivered in the Internet to rel.
     */ 
    public static void setReliability(double rel) {
        reliability = rel;
    }  // public static void setReliability(double)

    /**
     *  Internet.send(pkt) attempts delivery of pkt to its destination.
     *  The probability of its arrival is Internet.getReliability().
     *  The average round-trip time is Internet.getRoundTripTime().
     *  Mysteriously, CLOSE_TYPE packets always arrive and their
     *  round-trip time is always about 1.5*max_rtt.
     */ 
    public static void send(final Packet pkt) {
        Logger.log("Packet sent", pkt);
        boolean isClosePacket = pkt.getType() == Packet.CLOSE_TYPE;
        boolean deliver = isClosePacket || shouldDeliver();
        if (deliver) {
            int delay;
            if (isClosePacket) {
                delay = Math.round((float)(1.5*max_rtt));
            } else {
                delay = getDelay();
            }
            TimerFactory.Timer timer = timerFactory.createTimer(delay,
                    new TimingTask() {
                public void doTask(Packet pkt) {
                    Logger.log("Packet delivered", pkt);
                    pkt.getDestination().receivePacket(pkt);
                }
            });
            timer.setPacket(pkt.getCopy());
            timer.start();
        }
    }  // public static void send(final Packet)

}  // public class Internet

