package rdt;

/**
 *  <p>
 *  A Packet is a protocol data unit used at the interface between a
 *  TCPEntity and an Internet.
 *  A Packet has a type, identification number, sequence number, source
 *  TCPEntity, and destination TCPEntity.
 *  All of these attributes except for the identification number and the
 *  sequence number are specified when the Packet is constructed and
 *  cannot be changed.
 *  </p>
 *  <p>
 *  The sequence number is normally assigned by the send side of a
 *  reliable data transfer protocol when it is initialized.
 *  The identification number can be set to simulate filling the packet
 *  with data, allowing a fixed set of packets to be reused for the
 *  duration of a connection.
 *  </p>
 */
public class Packet {

    /**
     *  Packet.DATA_TYPE is the type for data packets.
     */   
    public static final String DATA_TYPE = "DATA";

    /**
     *  Packet.ACK_TYPE is the type for acknowledgement packets.
     */   
    public static final String ACK_TYPE = "ACK";

    /**
     *  Packet.NACK_TYPE is the type for negative acknowledgement packets.
     */   
    public static final String NACK_TYPE = "NACK";

    /**
     *  Packet.CLOSE_TYPE is the type for packets that close a connection.
     */   
    public static final String CLOSE_TYPE = "CLOSE";

    /**
     *  type is the type of this Packet.
     */   
    private String type;

    /**
     *  identificationNumber is the identification number for this Packet.
     */   
    private int identificationNumber;

    /**
     *  source is the source TCPEntity for this Packet.
     */   
    private TCPEntity source;

    /**
     *  destination is the destination TCPEntity for this Packet.
     */   
    private TCPEntity destination;

    /**
     *  sequenceNumber is the sequence number for this Packet.
     */   
    private int sequenceNumber;

    /**
     *  This constructor is disabled.
     *  A Packet must have a source and a destination.
     */ 
    private Packet() {
    }  // private Packet()

    /**
     *  new Packet(src, dest) returns a new Packet.
     *  Its source TCPEntity is src and its destination TCPEntity is dest.
     *  The sequence number of the new Packet should be assigned by its
     *  source RDTProtocol.
     */ 
    public Packet(TCPEntity src, TCPEntity dest) {
        source = src;
        destination = dest;
    }  // public Packet(TCPEntity, TCPEntity)

    /**
     *  pkt.getCopy() returns a clone of pkt.
     */ 
    public Packet getCopy() {
        Packet copy = new Packet(source, destination);
        copy.type = type;
        copy.identificationNumber = identificationNumber;
        copy.sequenceNumber = sequenceNumber;
        return copy;
    }  // public Packet getCopy()

    /**
     *  pkt.getType() returns the type of pkt.
     */   
    public String getType() {
        return type;
    }  // public String getType()

    /**
     *  pkt.setType(typ) sets the type of pkt to typ.
     *  The parameter typ should be one of the constants Packet.DATA_TYPE,
     *  Packet.ACK_TYPE, Packet.NACK_TYPE, or Packet.CLOSE_TYPE.
     */   
    public void setType(String typ) {
        type = typ;
    }  // public void setType(String)

    /**
     *  pkt.getIdentificationNumber() returns the identification number
     *  of pkt.
     */   
    public int getIdentificationNumber() {
        return identificationNumber;
    }  // public int getIdentificationNumber()

    /**
     *  pkt.setIdentificationNumber(id) sets the identification number of
     pkt to id.
     */   
    public void setIdentificationNumber(int id) {
        identificationNumber = id;
    }  // public void setIdentificationNumber(int)

    /**
     *  pkt.getSequenceNumber() returns the sequence number of pkt.
     */   
    public int getSequenceNumber() {
        return sequenceNumber;
    }  // public int getSequenceNumber()

    /**
     *  pkt.setSequenceNumber(seq) sets the sequence number of pkt to seq.
     */   
    public void setSequenceNumber(int seq) {
        sequenceNumber = seq;
    }  // public void setSequenceNumber(int)

    /**
     *  pkt.getSource() returns the source TCPEntity of pkt.
     */   
    public TCPEntity getSource() {
        return source;
    }  // public TCPEntity getSource()

    /**
     *  pkt.getDestination() returns the destination TCPEntity of pkt.
     */   
    public TCPEntity getDestination() {
        return destination;
    }  // public TCPEntity getDestination()

    /**
     *  pkt.toString() returns a String describing pkt.
     *  This String has the form
     *  type.identificationNumber.sequenceNumber.source->destination.
     *  The identification number is preceeded by the letter "I" and
     *  the sequence number is preceeded by the letter "S".
     */   
    public String toString() {
        return type + ".I" + identificationNumber + ".S" + sequenceNumber +
                "." + source + "->" + destination;
    }  // public String toString();

}  // public class Packet

