package rdt;

/**
 *  <p>
 *  An RDTProtocol implements a reliable data transfer protocol on
 *  behalf of a TCP connection thread in the Transport layer.
 *  It operates within a context provided by a TCP entity object, which
 *  provides a sandbox environment that hides synchronization with the
 *  application layer above and the network layer below.
 *  Each RDTProtocol is associated with a specific socket on a specific
 *  host.
 *  </p>
 *  <p>
 *  When created, a RDTProtocol is given a name.
 *  After setup, it has a partner, which is the RDTProtocol at the other
 *  end of the socket connection.
 *  In reality, of course, setup would only give it the host internet ID
 *  and port number of the partner.
 *  </p>
 *  <h3> Implementing Subclasses </h3>
 *  <p>
 *  To implement a reliable data transfer protocol, write a class that
 *  extends the RDTProtocol class.
 *  This class should provide three kinds of functionality:
 *  <ul>
 *  <li> initialization
 *  <li> receive side
 *  <li> send side
 *  </ul>
 *  </p>
 *  <p>
 *  Initialization is provided by declaring variables needed for receive
 *  and send side functionality and implementing the
 *  <code>doInit()</code> abstract method.
 *  The createPacket() and createTimer() methods can be used to
 *  construct packets and timers needed by the protocol.
 *  The setAcceptingNewSegments() and setPendingAcknowledgements()
 *  methods can be used to initalize the TCPEntity.
 *  </p>
 *  <p>
 *  Receive side functionality is provided by implementing the following
 *  abstract methods.
 *  <ul>
 *  <li> handleReceivedData()
 *  <li> handleReceivedAck()
 *  </ul>
 *  In addition, the following methods may be overridden if the default
 *  implementation is not acceptable.
 *  <ul>
 *  <li> handleReceivedNack()
 *  <li> handleReceivedClose()
 *  </ul>
 *  All of these methods are invoked by the demultiplexer in the
 *  TCPEntity in response to received packets.
 *  </p>
 *  <p>
 *  When a received segment is ready it can be sent to the reassembler
 *  in the TCPEntity using the reassemble() method.
 *  The terminate() method can be used to terminate the thread used by
 *  the protocol in response to a received close packet.
 *  </p>
 *  <p>
 *  Send side functionality is provided by implementing the following
 *  abstract methods.
 *  <ul>
 *  <li> handleNewSegment()
 *  <li> handleTimedOutPacket()
 *  </ul>
 *  In addition, the following method may be overridden if the default
 *  implementation is not acceptable.
 *  <ul>
 *  <li> initiateDisconnect()
 *  </ul>
 *  The <code>handleNewSegment()</code> method is invoked by the
 *  segmenter in the TCPEntity in response to messages from the
 *  application layer.
 *  The <code>handleTimedOutPacket()</code> method is invoked by the
 *  TCPEntity in response to expiration of a timer.
 *  The <code>initiateDisconnect()</code> method is invoked by the
 *  TCPEntity in response to a close operation from the application
 *  layer.
 *  </p>
 *  <p>
 *  The send() method can be used for passing packets to the transport
 *  layer multiplexer, which will send them to the other end of the
 *  connection.
 *  The setAcceptingNewSegments() method can be used to notify the
 *  TCPEntity whether or not new segments are being accepted.
 *  The setPendingAcknowledgements() method can be used to notify the
 *  TCPEntity whether or not there are packets that have been sent but not
 *  acknowledged.
 *  This controls when the TCPEntity will begin a disconnect.
 *  </p>
 */
public abstract class RDTProtocol {

    /**
     *  context provides access to services from the TCPEntity that is
     *  using this RDTProtocol.
     */   
    private TCPEntity.Context context;

    /**
     *  createPacket() returns a new Packet that only requires setting
     *  its type, its identification number, and its sequence number
     *  before use.
     */   
    protected Packet createPacket() {
        return context.createPacket();
    }  // protected Packet createPacket()

    /**
     *  createTimer() returns a new Timer that only requires setting
     *  its packet before use.
     */   
    protected RDTTimer createTimer() {
        return context.createTimer();
    }  // protected RDTTimer createTimer()

    /**
     *  setAcceptingNewSegments(acc) informs the TCPEntity whether or
     *  not this RDTProtocol is accepting new segments.
     *  <p>
     *  If the TCPEntity is not properly informed, the protocol may
     *  hang or behave unpredictably.
     */   
    protected void setAcceptingNewSegments(boolean acc) {
        context.setAcceptingNewSegments(acc);
    }  // protected void setAcceptingNewSegments(boolean)

    /**
     *  setPendingAcknowledgements(pa) informs the TCPEntity whether
     *  or not this RDTProtocol has packets that have been sent but
     *  not yet acknowledged.
     *  <p>
     *  If the TCPEntity is not properly informed, the protocol may
     *  close prematurely or not close at all.
     */   
    protected void setPendingAcknowledgements(boolean pa) {
        context.setPendingAcknowledgements(pa);
    }  // protected void setPendingAcknowledgements(boolean)

    /**
     *  reassemble(id) sends id to the reassembler.
     */   
    protected void reassemble(int id) {
        context.reassemble(id);
    }  // protected void reassemble(int)

    /**
     *  send(pkt) sends pkt to the receiving end of the connection.
     */   
    protected void send(Packet pkt) {
        Internet.send(pkt);
    }  // protected void send(Packet)

    /**
     *  terminate() terminates the thread used by this RDTProtocol.
     */   
    protected void terminate() {
        context.terminate();
    }  // protected void terminate()

    /**
     *  doInit() is invoked to perform protocol-dependent
     *  initialization for this RDTProtocol.
     */   
    protected abstract void doInit();

    /**
     *  rdtp.handleNewSegment(id) is invoked to get rdtp to handle a
     *  new segment with identification number id.
     *  The identification number is just a stand-in for the real
     *  data in a segment.
     *  It is used by the Logger for recording purposes.
     */   
    public abstract void handleNewSegment(int id);

    /**
     *  rdtp.handleReceivedData(pkt) is invoked to get rdtp to handle
     *  a DATA_TYPE packet received from its TCPEntity's connection
     *  partner.
     */   
    public abstract void handleReceivedData(Packet pkt);

    /**
     *  rdtp.handleReceivedAck(pkt) is invoked to get rdtp to handle
     *  a ACK_TYPE packet received from its TCPEntity's connection
     *  partner.
     */   
    public abstract void handleReceivedAck(Packet pkt);

    /**
     *  rdtp.handleTimedOutPacket(pkt) is invoked to get rdtp to
     *  handle a packet acknowledgement timeout for pkt.
     *  pkt may be null for protocols that use a single timer.
     */   
    public abstract void handleTimedOutPacket(Packet pkt);

    /**
     *  rdtp.handleReceivedNack(pkt) is invoked to get rdtp to
     *  handle a NACK_TYPE packet received from its TCPEntity's
     *  connection partner.
     *  The default implementation ignores NACK_TYPE packets.
     */   
    public void handleReceivedNack(Packet pkt) {
    }  // public void handleReceivedNack(Packet)

    /**
     *  rdtp.handleReceivedClose(pkt) is invoked to get rdtp to
     *  handle a CLOSE_TYPE packet received from its TCPEntity's
     *  connection partner.
     *  The default implementation terminates the connection.
     *  If pkt's sequence number is 0 then a CLOSE_TYPE packet
     *  with sequence number 1 is sent as acknowledgement before
     *  termination.
     */   
    public void handleReceivedClose(Packet pkt) {
        if (pkt.getSequenceNumber() == 0) {
            Packet ackPacket = createPacket();
            ackPacket.setType(Packet.CLOSE_TYPE);
            ackPacket.setSequenceNumber(1);
            send(ackPacket);
        }
        terminate();
    }  // public void handleReceivedClose(Packet)

    /**
     *  rdtp.initiateDisconnect() is invoked to get rdtp to initate
     *  a disconnect sequence.
     *  The defualt implementation sends a CLOSE_TYPE packet with
     *  sequence number 0 to the partner.
     */   
    public void initiateDisconnect() {
        Packet closePacket = createPacket();
        closePacket.setType(Packet.CLOSE_TYPE);
        closePacket.setSequenceNumber(0);
        send(closePacket);
    }  // public void initiateDisconnect()

    /**
     *  rdtp.initialize(cntxt) is invoked when a connection is set
     *  up using rdtp.
     *  Services from the TCPEntity are provided through cntxt.
     */   
    public void initialize(TCPEntity.Context cntxt) {
        context = cntxt;
        doInit();
    }  // public void initialize(TCPEntity.Context)

    /**
     *  rdtp.toString() returns the name of rdtp, which is derived
     *  from its TCPEntity's name.
     */   
    public String toString() {
        return context.getName() + "\'s protocol";
    }  // public String toString()

}  // public abstract class RDTProtocol

