package rdt;

/**
 *  <p>
 *  A TCPEntity represents a TCP connection thread in the Transport layer.
 *  Each TCPEntity is associated with a specific socket on a specific host.
 *  It delegates packet handling to a reliable data transfer protocol.
 *  </p>
 *  <p>
 *  When created, a TCPEntity is given a name.
 *  When setup, it is given a partner, which is the TCPEntity at the other
 *  end of the socket connection, and an RDTProtocol, which specifies the
 *  reliable data transfer protocol that it should use.
 *  In reality, of course, the partner TCPEntity would be replaced by just
 *  the host internet ID and port number of the partner.
 *  </p>
 *  <h3> Internal View </h3>
 *  <p>
 *  A TCPEntity provides a context in which reliable data transfer
 *  protocol code can be executed in a single thread.
 *  It provides all necessary synchronization with other threads.
 *  It also, through its nested Context class, provides services that the
 *  reliable data transfer protocol can use to control the TCPEntity and
 *  to interface with the demultiplexer and reassembler on the receive
 *  side, and the segmenter and multiplexer on the send side.
 *  </p>
 */
public class TCPEntity
implements Runnable {

    /**
     *  packetPrototype is a packet prototype for this TCPEtity.
     *  It can be cloned to create new packets.
     */   
    private Packet packetPrototype;

    /**
     *  timerPrototype is a timer prototype for this TCPEtity.
     *  It can be cloned to create new timers.
     */   
    // private RDTTimer timerPrototype;

    /**
     *  timerFactory is a timer factory for this TCPEtity.
     *  It can be used to create new timers.
     */   
    private TimerFactory timerFactory;

    /**
     *  timerDelay is the delay for timers for this TCPEtity.
     */   
    private int timerDelay;

    /**
     *  addTimedOutPacketTask is the timing task to be performed by a
     *  timer when a packet has timed out waiting for acknowledgement.
     */   
    private TimingTask addTimedOutPacketTask;

    /**
     *  acceptingNewSegments is true when protocol is accepting new
     *  segments.
     */   
    private boolean acceptingNewSegments;

    /**
     *  pendingAcknowledgements is true when protocol has packets that have
     *  been sent but not yet acknowledged.
     */   
    private boolean pendingAcknowledgements;

    /**
     *  reassembler reassembles the message received by this TCPEntity.
     */   
    private StringBuffer reassembler;

    /**
     *  connected is true if this TCPEntity is connected to its
     *  communication partner.
     */   
    private boolean connected;

    /**
     *  name is the name of this TCPEntity.
     */   
    private String name;

    /**
     *  A Context is created by a TCPEntity and passed to its RDTProtocol
     *  during set up of the TCPEntity.
     *  The Context gives the RDTProtocol access to private data of the
     *  TCPEntity.
     *  After setUp() is invoked, the data accessed through the Context is 
     *  not accessed except on the protocol thread so the access can be
     *  unsynchronized.
     */   
    public class Context {

        /**
         *  cntxt.createPacket() returns a new packet that only requires
         *  setting its identification number an sequence number befor use.
         */   
        public Packet createPacket() {
            return packetPrototype.getCopy();
        }  // public Packet createPacket()

        /**
         *  cntxt.createTimer() returns a new timer that only requires
         *  setting its packet before use.
         */   
        public RDTTimer createTimer() {
            return timerFactory.createTimer(timerDelay, addTimedOutPacketTask);
        }  // public RDTTimer createTimer()

        /**
         *  cntxt.setAcceptingNewSegments(ans) allows this TCPEntity to
         *  schedule new segment tasks if ans is true.
         */   
        public void setAcceptingNewSegments(boolean ans) {
            acceptingNewSegments = ans;
        }  // public void setAcceptingNewSegments(boolean)

        /**
         *  cntxt.setPendingAcknowledgements(pa) notifies this TCPEntity
         *  that its protocol has packets that have been sent but not yet
         *  acknowledged.
         */   
        public void setPendingAcknowledgements(boolean pa) {
            pendingAcknowledgements = pa;
        }  // public void setPendingAcknowledgements(boolean)

        /**
         *  cntxt.reassemble(id) sends the segment id to the reassembler
         *  of this TCPEntity.
         */   
        public void reassemble(int id) {
            reassembler.append("Segment " + id + "\n");
        }  // public void reassemble(int)

        /**
         *  cntxt.terminate() terminates the execution of the thread
         *  created by this TCPEntity.
         */   
        public void terminate() {
            connected = false;
        }  // public void terminate()

        /**
         *  cntxt.getName() returns the name of the TCPEntity that created
         *  this Context.
         */   
        public String getName() {
            return name;
        }  // public String getName()

    }  // public class Context

    /**
     *  protocol is the RDTProtocol used by this TCPEntity.
     *  <p>
     *  After setUp() is invoked, this variable is only accessed on the 
     *  protocol thread.
     */   
    private RDTProtocol protocol;

    /**
     *  disconnecting is true if this TCPEntity has initiated a disconnect.
     *  <p>
     *  This variable is modified by the protocol thread and read by other
     *  threads.
     *  All access must be synchronized.
     */   
    private boolean disconnecting;

    /**
     *  protocolSleeping is true when the protocol Thread is waiting for a
     *  task.
     *  <p>
     *  This variable is modified by the protocol thread and read by other
     *  threads.
     *  All access must be synchronized.
     */   
    private boolean protocolSleeping;

    /**
     *  getTaskCount(ans) returns the number of scheduled tasks in the task
     *  queues for this TCPEntity.
     *  New segment tasks are counted only if ans is true.
     *  <p>
     *  This method is only executed on the protocol thread.
     */
    private int getTaskCount() {
        if (!connected) {
            return 1;
        }
        int taskCount = receivedPacketQueue.getCount();
        taskCount += timedOutPacketQueue.getCount();
        if (acceptingNewSegments) {
            taskCount += sendQueue.getCount();
        }
        if (sendQueue.getCount() == 0 && disconnecting &&
                !pendingAcknowledgements) {
            taskCount++;
        }
        return taskCount;
    }  // private int getTaskCount()

    /**
     *  doSelectedTask() performs a task from one of the task queues.
     *  Modify this method to change the relative priorities of the
     *  different kinds of tasks.
     *  Just change the order of the if statements.
     *  <p>
     *  This method is only executed on the protocol thread.
     */
    private void doSelectedTask() {
        if (!connected) {
            return;
        }
        if (sendQueue.getCount() > 0 && acceptingNewSegments) {
            int id = sendQueue.get();
            sendQueue.remove();
            protocol.handleNewSegment(id);
            return;
        }
        if (receivedPacketQueue.getCount() > 0) {
            Packet packet = receivedPacketQueue.get();
            receivedPacketQueue.remove();
            if (packet.getType() == Packet.DATA_TYPE) {
                protocol.handleReceivedData(packet);
            } else if (packet.getType() == Packet.ACK_TYPE) {
                protocol.handleReceivedAck(packet);
            } else if (packet.getType() == Packet.NACK_TYPE) {
                protocol.handleReceivedNack(packet);
            } else if (packet.getType() == Packet.CLOSE_TYPE) {
                protocol.handleReceivedClose(packet);
            }
            return;
        }
        if (timedOutPacketQueue.getCount() > 0) {
            Packet packet = timedOutPacketQueue.get();
            timedOutPacketQueue.remove();
            protocol.handleTimedOutPacket(packet);
            return;
        }
        // This should be the last check.
        if (sendQueue.getCount() == 0 && disconnecting &&
                !pendingAcknowledgements) {
            protocol.initiateDisconnect();
            disconnecting = false;
            return;
        }
        Logger.log("Null task for " + this);
    }  // private void doSelectedTask()

    /**
     *  doTask() performs the next scheduled task in the task queue for
     *  the protocol.
     *  This method blocks until there is a scheduled task.
     *  <p>
     *  This method is only executed on the protocol thread.
     */
    private synchronized void doTask() {
        // First wait until there is a task.
        protocolSleeping = true;
        while (getTaskCount() == 0) {
            try {
                wait();
            } catch(InterruptedException e) {
                e.printStackTrace(System.err);
                System.exit(1);
            }
        }
        protocolSleeping = false;
        doSelectedTask();
    }  // private synchronized void doTask()

    /**
     *  sendQueue is the queue of segments to be sent by this TCPEntity.
     *  <p>
     *  This variable is modified by the protocol thread and read by other
     *  threads.
     *  All access must be synchronized.
     */   
    private IntQueue sendQueue;

    /**
     *  receivedPacketQueue is the queue of received packets for this
     *  TCPEntity.
     *  <p>
     *  This variable is modified by the protocol thread and read by other
     *  threads.
     *  All access must be synchronized.
     */   
    private PacketQueue receivedPacketQueue;

    /**
     *  timedOutPacketQueue is the queue of timed out packets for this
     *  TCPEntity.
     *  <p>
     *  This variable is modified by the protocol thread and read by other
     *  threads.
     *  All access must be synchronized.
     */   
    private PacketQueue timedOutPacketQueue;

    /**
     *  tcpe.addTimedOutPacket(pkt) is invoked to get tcpe to handle a packet
     *  time out.
     */   
    private synchronized void addTimedOutPacket(Packet pkt) {
        // Logger.log("addTimedOutPacket()", pkt);
        timedOutPacketQueue.add(pkt);
        if (protocolSleeping) {
            notifyAll();
        }
    }  // private synchronized void addTimedOutPacket(Packet)

    /**
     *  This constructor is disabled.
     *  A TCPEntity must have a name.
     */ 
    private TCPEntity() {
    }  // private TCPEntity()

    /**
     *  new TCPEntity(nm) returns a new TCPEntity named nm.
     */ 
    public TCPEntity(String nm) {
        name = nm;
        sendQueue = new IntQueue();
        receivedPacketQueue = new PacketQueue();
        timedOutPacketQueue = new PacketQueue();
        reassembler = new StringBuffer();
        addTimedOutPacketTask = new TimingTask() {
            public void doTask(Packet pkt) {
                addTimedOutPacket(pkt);
            }
        };
    }  // public TCPEntity(String, RDTProtocol)

    /**
     *  tcpe.setUp(prtnr, proto) sets up tcpe for communication with prtnr
     *  using the reliable data transfer protocol proto.
     *  The Thread used by tcpe is returned.
     */   
    public Thread setUp(TCPEntity prtnr, RDTProtocol proto) {
        // FIXME - should add methods to clear queues and invoke them here.
        protocol = proto;
        Thread thread = new Thread(this);
        reassembler.setLength(0);
        connected = true;
        disconnecting = false;
        packetPrototype = new Packet(this, prtnr);
        timerDelay = Internet.getRoundTripTime()*5/4;
        timerFactory = new TimerFactory(toString());
        Context context = new Context();
        protocol.initialize(context);
        thread.start();
        return thread;
    }  // public Thread setUp(TCPEntity)

    /**
     *  tcpe.sendData(id) sends a new data packet with identification
     *  number id from tcpe to its communication partner.
     *  <p>
     *  This method is the segmenter's interface to the TCPEntity.
     *  It is invoked to schedule a segment for packetization and delivery.
     *  The id parameter is just a stand-in for the real data in a segment.
     *  It is used by the Logger for recording purposes.
     */   
    public synchronized void sendData(int id) {
        // Logger.log("sendData(" + id + ")");
        sendQueue.add(id);
        if (protocolSleeping) {
            notifyAll();
        }
    }  // public synchronized void sendData(int)

    /**
     *  tcpe.receivePacket(pkt) is invoked to get tcpe to handle a packet
     *  from its connection partner.
     *  <p>
     *  This method is the demultiplexer's interface to the TCPEntity.
     */   
    public synchronized void receivePacket(Packet pkt) {
        // Logger.log("receivePacket()", pkt);
        receivedPacketQueue.add(pkt);
        if (protocolSleeping) {
            notifyAll();
        }
    }  // public synchronized void receivePacket(Packet)

    /**
     *  tcpe.getReceivedMessage() returns the message received by tcpe.
     */   
    public String getReceivedMessage() {
        return reassembler.toString();
    }  // public String getReceivedMessage()

    /**
     *  tcpe.disconnect() is invoked to get tcpe to disconnect from its
     *  communication partner.
     *  The disconnect is only done after pending tasks have been done.
     */   
    public synchronized void disconnect() {
        // Logger.log("Initiating disconnect");
        disconnecting = true;
        if (protocolSleeping) {
            notifyAll();
        }
    }  // public synchronized void disconnect()

    /**
     *  tcpe.toString() returns the name of tcpe.
     */   
    public String toString() {
        return name;
    }  // public String toString()

    /**
     *  Do not invoke this method externally - it is automatically invoked
     *  at the end of the setUp() method.
     *  <p>
     *  tcpe.run() repeatedly looks for tasks from the scheduler and
     *  executes them.
     *  This method is the top-level code executed by the protocol thread.
     *  When the method terminates, the thread terminates.
     */   
    public void run() {
        while (connected) {
            doTask();
        }
    }  // public void run()

}  // public class TCPEntity

