package rdt;

import java.util.*;

/**
 *  A TimerFactory is a factory for RDTTimer objects.
 *  It provides a single thread for executing the tasks for its timers.
 */
public class TimerFactory {

    /**
     *  STOPPED is the index used for all stopped Timer objects.
     */
    private static final int STOPPED = -1;

    /**
     *  name is the name of this TimerFactory.
     */
    private String name;

    /**
     *  timerCount is the number of Timer objects created by this
     *  TimerFactory.
     */
    private int timerCount;

    /**
     *  A TimerFactory.Timer times tasks that need to be performed at a
     *  later time.
     *  <p>
     *  A TimerFactory.Timer has a task and a delay.
     *  The delay determines the interval of time between starting the
     *  timer and performing the task.
     *  A TimerFactory.Timer is is either an active state or a stopped state.
     *  When it is in an active state it has a scheduled time, which is the
     *  time when its task will be performed.
     *  When it is in a stopped state its task has no scheduled time.
     */
    public class Timer
    extends RDTTimer {

	/**
	 *  delay is the time that this Timer waits before performing
	 *  its task.
	 */
	private long delay;

	/**
	 *  task is the task performed by this Timer.
	 */
	private TimingTask task;

	/**
	 *  index is the index of this Timer in its TimerFactory.
	 *  If index is STOPPED then this Timer is stopped and it does
	 *  not have an index.
	 */
	private int index;

	/**
	 *  identifier is a number for distinguishing this Timer object
	 *  from other Timer objects.
	 */
	private int identifier;

	/**
	 *  scheduledTime is the time that this Timer's task should be
	 *  performed.
	 *  If this Timer is stopped then scheduledTime is ignored.
	 */
	private long scheduledTime;

	/**
	 *  new Timer(del, tsk) returns a new Timer that performs task tsk.
	 *  Its delay is del.
	 */
	public Timer(long del, TimingTask tsk) {
	    delay = del;
	    task = tsk;
	    index = STOPPED;
	    identifier = timerCount;
	    timerCount++;
	}  // public Timer(long, TimingTask)

	/**
	 *  tmr.start() sets the scheduled time for tmr's task to the
	 *  current time plus tmr's delay.
	 */
	public void start() {
	    startTimer(this, delay);
	}  // public void start()

	/**
	 *  tmr.start(del) sets the scheduled time for tmr's task to the
	 *  current time plus del.
	 */
	public void start(long del) {
	    startTimer(this, del);
	}  // public void start(long)

	/**
	 *  tmr.stop() puts tmr in the stopped state.
	 */
	public void stop() {
	    stopTimer(this);
	}  // public void stop()

	/**
	 *  tmr.toString() returns a String of the form name"-Timer-"id
	 *  where name is the name of the TimerFactory that created tmr and
	 *  id is the identification number of tmr.
	 */
	public String toString() {
	    return name + "-Timer-" + identifier;
	}  // public String toString()

    }  // public class Timer

    /**
     *  activeTimers is the list of active Timer objects in this
     *  TimerFactory.
     */
    private List activeTimers;

    /**
     *  get() returns the timer with the earliest scheduled time in
     *  activeTimers.
     *  <p>
     *  Precondition:
     *      activeTimers.size() > 0
     */
    private Timer get() {
	// Check the precondition and throw an exception if it is false.
	if (activeTimers.size() <= 0) {
	    throw new IllegalStateException("get(): activeTimers is empty.");
	}
	// Return the timer at the beginning of the heap.
	return (Timer)activeTimers.get(0);
    }  // private Timer get()

    /**
     *  trickleUp(iViol) restores the heap condition in activeTimers.
     * 
     *  Precondition:
     *      activeTimers satisfies the heap condition except for a
     *      possible violation between the timer at iViol and its parent.
     */
    private void trickleUp(int iViol) {
	int iParent;		// index of parent
	Timer parent;		// parent timer
	Timer viol;		// timer at iViol

	// If iViol is the heap root then return.
	if (iViol <= 0) {
	    return;
	}
	// If the timer at iViol is not scheduled before its parent timer
	// then return.
	iParent = (iViol - 1)/2;
	viol = (Timer)activeTimers.get(iViol);
	parent = (Timer)activeTimers.get(iParent);
	if (viol.scheduledTime >= parent.scheduledTime) {
	    return;
	}
	// Swap timers at iViol and iParent.  This may cause a new heap
	// violation between iParent and its parent.
	activeTimers.set(iViol, parent);
	parent.index = iViol;
	activeTimers.set(iParent, viol);
	viol.index = iParent;
	// Fix the possible new heap violation.
	trickleUp(iParent);
    }  // private void trickleUp(int)

    /**
     *  trickleDown(iViol) restores the heap condition in activeTimers.
     * 
     *  Precondition:
     *      activeTimers satisfies the heap condition except for a
     *      possible violation between the timer at iViol and its children.
     */
    private void trickleDown(int iViol) {
	int iLeft;		// index of the left child
	int iRight;		// index of the right child
	int iHigh;		// index of the higher priority child
	Timer viol;		// timer at iViol
	Timer leftChild;	// left child timer
	Timer rightChild;	// right child timer
	Timer highChild;	// higher priority child timer

	// If iViol has no children then return.
	iLeft = 2*iViol + 1;
	if (iLeft >= activeTimers.size()) {
	    return;
	}
	// Determine higher priority child and its index.
	iRight = iLeft + 1;
	leftChild = (Timer)activeTimers.get(iLeft);
	iHigh = iLeft;
	highChild = leftChild;
	if (iRight < activeTimers.size()) {
	    rightChild = (Timer)activeTimers.get(iRight);
	    if (rightChild.scheduledTime < leftChild.scheduledTime) {
		iHigh = iRight;
		highChild = rightChild;
	    }
	}
	// If higher priority child is not prior to timer at iViol then
	// return.
	viol = (Timer)activeTimers.get(iViol);
	if (highChild.scheduledTime >= viol.scheduledTime) {
	    return;
	}
	// Swap timers at iViol and highChild.  This may create a new heap
	// violation between iHigh and its children.
	activeTimers.set(iViol, highChild);
	highChild.index = iViol;
	activeTimers.set(iHigh, viol);
	viol.index = iHigh;
	// Repair the possible new heap violation.
	trickleDown(iHigh);
    }  // private void trickleDown(int)

    /**
     *  add(tmr) adds tmr to activeTimers.
     */
    private void add(Timer tmr) {
	// Add tmr at the end of the heap.  This may cause a heap violation
	// between the the last index and its parent.
	activeTimers.add(tmr);
	tmr.index = activeTimers.size() - 1;
	// Repair the possible new heap violation.
	trickleUp(activeTimers.size() - 1);
    }  // private void add(Timer)

    /**
     *  remove() removes the timer with the earliest scheduled time from
     *  activeTimers.
     *  <p>
     *  Precondition:
     *      activeTimers.size() > 0
     */
    private void remove() {
	remove(0);
    }  // private void remove()

    /**
     *  remove(indx) removes the timer with index indx from activeTimers.
     *  <p>
     *  Precondition:
     *      0 <= indx < activeTimers.size()
     */
    private void remove(int indx) {
	// Check the precondition and throw an exception if it is false.
	if (indx < 0 || activeTimers.size() <= indx) {
	    throw new IllegalStateException("remove(): invalid remove index.");
	}
	// Stop the timer at index indx in the activeTimers heap.
	((Timer)activeTimers.get(indx)).index = STOPPED;
	// Remove the last timer in activeTimers heap.
	Timer lastTimer = (Timer)activeTimers.remove(activeTimers.size() - 1);
	// If indx == activeTimers.size() then we don't need any
	// adjustment of the activeTimers heap.
	if (indx < activeTimers.size()) {
	    // Place lastTimer at index indx of activeTimers.  This may
	    // result in a heap violation between index indx and its
	    // children.
	    activeTimers.set(indx, lastTimer);
	    lastTimer.index = indx;
	    // Repair the possible new heap violation.
	    trickleDown(indx);
	}
    }  // private void remove(int)

    /**
     *  startTimer(tmr, del) starts timer tmr with a delay del.
     */
    private synchronized void startTimer(Timer tmr, long del) {
	long oldScheduledTime = tmr.scheduledTime;
	tmr.scheduledTime = System.currentTimeMillis() + del;
	int index = tmr.index;
	if (index == STOPPED) {
	    add(tmr);
	} else if (tmr.scheduledTime < oldScheduledTime) {
	    trickleUp(index);
	} else {
	    trickleDown(index);
	}
	// Whenever tmr is the next timer to be scheduled, wake up the
	// timer thread.
	if (tmr.index == 0) {
	    notifyAll();
	}
    }  // private synchronized void startTimer(Timer, long)

    /**
     *  stopTimer(tmr) stops timer tmr.
     */
    private synchronized void stopTimer(Timer tmr) {
	int index = tmr.index;
	if (index != STOPPED) {
	    remove(index);
	}
	// Whenever there are no active timers, wake up any thread that
	// may be waiting for all timers to be stopped.
	if (activeTimers.size() == 0) {
	    notifyAll();
	}
    }  // private synchronized void stopTimer(Timer)

    /**
     *  timerThread is the Thread in which timer tasks for this
     *  TimerFactory are executed.
     */
    private Thread timerThread;

    /**
     *  doCurrentTasks() executes all active tasks whose scheduled time is
     *  at or before the current time.
     *  Then the invoking thread blocks until the scheduled time for the
     *  next active task or until notified of a change in the next
     *  scheduled time.
     */
    private synchronized void doCurrentTasks() {
	long waitTime = 0;
	boolean currentTasksDone = false;
	while (!currentTasksDone) {
	    if (activeTimers.size() == 0) {
		waitTime = 0;
		currentTasksDone = true;
	    } else {
		Timer nextTimer = get();
		waitTime = nextTimer.scheduledTime - System.currentTimeMillis();
		currentTasksDone = waitTime > 0;
		if (!currentTasksDone) {
		    stopTimer(nextTimer);
		    nextTimer.task.doTask(nextTimer.getPacket());
		}
	    }
	}
	try {
	    wait(waitTime);
	} catch(InterruptedException e) {
	    e.printStackTrace();
	}
    }  // private synchronized void doCurrentTasks()

    /**
     *  This constructor is disabled.
     *  A TimerFactory must have a name.
     */
    private TimerFactory() {
    }  // TimerFactory()

    /**
     *  TimerFactory(nm) returns an empty TimerFactory with name nm.
     */
    public TimerFactory(String nm) {
	name = nm;
	activeTimers = new ArrayList();
	timerThread = new Thread(new Runnable() {
	    public void run() {
		while (true) {
		    doCurrentTasks();
		}
	    }
	});
	timerThread.setDaemon(true);
	timerThread.start();
    }  // TimerFactory(String)

    /**
     *  tq.createTimer(del, tsk) returns a Timer for use in tq that
     *  performs tsk with delay del after invoking start().
     */
    public Timer createTimer(int del, TimingTask tsk) {
	return new Timer(del, tsk);
    }  // public Timer createTimer(int, TimingTask)

    /**
     *  tq.waitUntilStopped() blocks until all timers are stopped.
     */
    public synchronized void waitUntilStopped() {
	while (activeTimers.size() > 0) {
	    try {
		wait();
	    } catch (InterruptedException e) {
		e.printStackTrace();
	    }
	}
    }  // public synchronized void waitUntilStopped()

    /**
     *  tq.toString() returns the name of tq.
     */
    public String toString() {
	return name;
    }  // public String toString()

}  // public class TimerFactory

